dnl Was in Aurelia, and modified since.

dnl This file is a part of Aurelia.
dnl Copyright (C) 2010, 2011  Valentin David
dnl Copyright (C) 2010  University of Bergen
dnl
dnl This program is free software: you can redistribute it and/or modify
dnl it under the terms of the GNU General Public License as published by
dnl the Free Software Foundation, either version 3 of the License, or
dnl (at your option) any later version.
dnl
dnl This program is distributed in the hope that it will be useful,
dnl but WITHOUT ANY WARRANTY; without even the implied warranty of
dnl MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
dnl GNU General Public License for more details.
dnl
dnl You should have received a copy of the GNU General Public License
dnl along with this program.  If not, see <http://www.gnu.org/licenses/>.

AC_DEFUN([NOTCXX11], [
  AC_MSG_ERROR([$1. Is $CXX a C++11 compiler?])
])

AC_DEFUN([NOTCXX03], [
  AC_MSG_ERROR([$1. Is $CXX standard compliant?])
])

AC_DEFUN([CHECK_CXX11_HEADER], [
  AC_LANG_PUSH([C++])
  AC_CHECK_HEADER([$1],[],[
    NOTCXX11([Header $1 set is missing])
  ])
  AC_LANG_POP
])

AC_DEFUN([CHECK_CXX03_HEADER], [
  AC_LANG_PUSH([C++])
  AC_CHECK_HEADER([$1],[],[
    NOTCXX03([Header $1 is missing])
  ])
  AC_LANG_POP
])

AC_DEFUN([CHECK_CXX11_HEADERS], [
  AH_CHECK_HEADERS([$1])
  m4_map_args_w([$1],[CHECK_CXX11_HEADER(],[)])
])

AC_DEFUN([CHECK_CXX03_HEADERS], [
  AH_CHECK_HEADERS([$1])
  m4_map_args_w([$1],[CHECK_CXX03_HEADER(],[)])
])

AC_DEFUN([CXX11_FUNCTION],[
AC_MSG_CHECKING([whether std::function exists])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[#include <functional>
struct Fun {
  int operator()(int) const {
    return 0;
  }
};
int test() {
  std::function<int(int)> f = Fun();
  return f(0);
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([std::function is missing])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_LAMBDAS],[
AC_MSG_CHECKING([whether the compiler supports lambda expressions])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[int test(int n) {
  return [] (int n) { return n; }(n);
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Lambda expression are not supported])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_VARIADIC_TEMPLATES],[
AC_MSG_CHECKING([whether the compiler supports variadic templates])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
int test() {
  return 0;
}
template <typename T, typename... U>
int test(T, U... u) {
  return 1 + test(u...);
}
]],[[test(1, 2, 3, 4);]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Variadic templates are not supported])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_TYPED_ENUMS],[
AC_MSG_CHECKING([whether enums can be typed])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
enum foo : unsigned { bar = 0 };
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Enumerations cannot be typed])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_AUTO],[
AC_MSG_CHECKING([whether keyword auto can be used as a type])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
int foo() {
  auto i = 0;
  return i;
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Keyword auto is not usable])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_DECLTYPE],[
AC_MSG_CHECKING([whether keyword decltype is supported])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
static int t;
decltype(t) foo() {
  return 0;
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Keyword decltype is not supported])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_DECLVAL],[
AC_MSG_CHECKING([whether std::declval is provided])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
#include <memory>
template <unsigned N>
struct A {};
A<sizeof(std::declval<int>())> a;
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([Function template std::declval is not supported])
])
AC_LANG_POP
])


AC_DEFUN([CXX11_STATIC_ASSERT],[
AC_MSG_CHECKING([whether static_assert is supported])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
void foo() {
  static_assert(true, "foo");
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_RESULT([no])
  NOTCXX11([static_asser is not supported])
])
AC_LANG_POP
])

AC_DEFUN([CXX11_FLAG],[
AC_MSG_CHECKING([whether the C++ compiler accepts -std=c++0x.])
AC_LANG_PUSH([C++])
oldCPPFLAGS="$CPPFLAGS"
CPPFLAGS="-std=c++0x $CPPFLAGS"
AC_COMPILE_IFELSE(
 [AC_LANG_SOURCE([[]])],
 [EXTRACPPFLAGS="-std=c++0x"
  AC_MSG_RESULT([yes])],
 [EXTRACPPFLAGS=""
  AC_MSG_RESULT([no])])
CPPFLAGS="$EXTRACPPFLAGS $oldCPPFLAGS"
AC_LANG_POP
])

AC_DEFUN([CXX_WALL],[
if ! test "$ac_test_CXXFLAGS" = set; then
  AC_LANG_PUSH([C++])
  oldCXXFLAGS="$CXXFLAGS"
  CXXFLAGS="$CXXFLAGS -W -Wall"
  AC_COMPILE_IFELSE([AC_LANG_PROGRAM()],[],[CXXFLAGS="$oldCXXFLAGS"])
  AC_LANG_POP
fi
])

AC_DEFUN([CXX_WERROR],[
if ! test "$ac_test_CXXFLAGS" = set; then
  AC_LANG_PUSH([C++])
  oldCXXFLAGS="$CXXFLAGS"
  CXXFLAGS="$CXXFLAGS -Werror"
  AC_COMPILE_IFELSE([AC_LANG_PROGRAM()],[],[CXXFLAGS="$oldCXXFLAGS"])
  AC_LANG_POP
fi
])

AC_DEFUN([CXX11_THREAD_LOCAL],[
AC_MSG_CHECKING([whether the compiler supports thread_local])
AC_LANG_PUSH([C++])
AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[int test() {
  thread_local int i = 10;
  return i;
}
]])],[AC_MSG_RESULT([yes])],[dnl
  AC_MSG_CHECKING([whether the compiler supports static __thread])
  AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[int test() {
    static __thread int i = 10;
    return i;
  }
  ]])],[dnl
  AC_MSG_RESULT([yes])
  AC_DEFINE([thread_local],[static __thread],[Replacement for thread_local])
  ],[dnl
    AC_MSG_RESULT([no])
    NOTCXX11([thread_local is not supported])
  ])
])
AC_LANG_POP
])

AC_DEFUN([CXX_SPECIAL_DEFINE],[dnl
  AC_MSG_CHECKING([whether $1 is defined])
  AC_LANG_PUSH([C++])
  AC_COMPILE_IFELSE([AC_LANG_PROGRAM([[
  void f($2);
  void g() { f($1); }
  ]])], [dnl
    AC_MSG_RESULT([yes])
    AC_DEFINE([DEFINE_$1],[1],[Tells whether $1 is defined])
  ],[dnl
    AC_MSG_RESULT([no])
  ])
])

