template <typename T, typename Op, typename Id>
struct monoid: public concept {
  typedef concept_list<
    is_callable<Op(T, T)>,
    is_callable<Id()>,
    std::is_convertible<typename is_callable<Op(T, T)>::result_type, T>,
    std::is_convertible<typename is_callable<Id()>::result_type, T>
    > requirements;

  static void associativity(const Op& op, const T& a,
		      const T& b, const T& c) {
    axiom_assert(op(a, op(b, c)) == op(op(a, b), c));
  }

  static void identity(const Op& op, const T& a, const Id& id) {
    axiom_assert((op(id(), a) == a) && (op(a, id()) == a));
  }

  AXIOMS(associativity, identity);
};

template <typename T>
struct monoid_plus: concept {
  typedef
    monoid<T, op_plus,
           wrapped_constructor<T()> >
    requirements;
};

namespace justtoshow {
struct op_plus {
  template <typename T, typename U,
	    typename Ret = decltype(std::declval<T>() + std::declval<U>())>
  Ret operator()(T&& t, U&& u) const {
    return std::forward<T>(t) + std::forward<U>(u);
  }
};

template <typename T, bool = is_constructible<T>::value>
struct wrapped_constructor;

template <typename T, typename... Args>
struct wrapped_constructor<T(Args...), true> {
  T operator()(Args... args...) const {
    return T(std::forward<Args>(args)...);
  }
};

template <typename T, typename... Args>
struct wrapped_constructor<T(Args...), false> {
};
}

namespace catsfoot {
  template <>
  struct verified<monoid<int, op_plus, wrapped_constructor<int()> > >
    : public std::true_type
  {};
}

namespace catsfoot {
  template <typename T>
  struct verified<monoid_plus<T> >
    : public verified<monoid<T, op_plus, wrapped_constructor<int()> > >
  {};
}
