// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <unordered_map>
#include <tuple>
# define __CATSFOOT_ENABLE_TESTING_RT 1
#include "axioms.hh"

namespace catsfoot {
  namespace details {
    typedef std::tuple<std::string, std::string, unsigned> axiom_id;

    struct hash_axiom {
      size_t operator()(const axiom_id& id) const {
	std::hash<std::string> sh;

	return sh(std::get<0>(id)) ^
	  sh(std::get<1>(id)) ^
	  std::get<2>(id);
      }
    };

    typedef std::unordered_map<axiom_id, unsigned, hash_axiom> rf_type;

    static rf_type reached_functions;

    void reached_function(std::string&& func,
			  std::string&& file,
			  unsigned line) {
      ++(reached_functions[axiom_id
			   (std::move(func), std::move(file), line)]);
    }

    void register_function(std::string&& func,
			   std::string&& file,
			   unsigned line) {
      reached_functions.insert(std::pair<axiom_id, unsigned>
			       {axiom_id(std::move(func),
					 std::move(file), line),
				   0});
    }
  }

  bool check_unverified(std::ostream& s) {
    bool ret = true;

    for (auto i = details::reached_functions.begin();
	 i != details::reached_functions.end(); ++i) {
      if (i->second == 0) {
	s << std::get<1>(i->first) << ":" << std::get<2>(i->first) << ":"
	  << " Axiom in " << std::get<0>(i->first)
	  << " has never been reached." << std::endl;
	ret = false;
      }
    }
    return ret;
  }

}
