// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef __TEST_ALL_DRIVER_HH
# define __TEST_ALL_DRIVER_HH

# include "test_driver.hh"

namespace catsfoot {
  namespace details {
    /// \brief Selected only when U has no static member get_axioms()
    template <typename U>
    std::false_type has_get_axiom_helper(try_second, U&&) {
      static_assert(details::always_false<U>::value,
		    "Cannot use this function");
      return 0;
    }

    /// \brief Selected only when U has a static member get_axioms()
    template <typename U,
	      typename =
	      decltype(U::get_axioms())>
    std::true_type has_get_axiom_helper(try_first, U&&) {
      static_assert(details::always_false<U>::value,
		    "Cannot use this function");
      return 0;
    }

    /// \brief Predicate testing that U has static member get_axioms()
    template <typename U>
    struct has_get_axiom:
      public identity
    <decltype(has_get_axiom_helper(try_first(),
				   std::declval<U>()))>
    ::type {
    };

    /// \brief Test axioms of a predicate (always true)
    template <typename T,
	      bool = is_concept<T>::value,
	      bool = has_get_axiom<T>::value>
    struct test_all {
      template <typename Generator,
		typename Stream = decltype(std::cerr)>
      bool operator()(Generator&, Stream& = std::cerr) {
	return true;
      }
    };

    /// \brief Test axioms of a concept with no local axioms
    template <typename T>
    struct test_all<T, true, false> {
      template <typename Generator,
	      typename Stream = decltype(std::cerr)>
      bool operator()(Generator& g, Stream& s = std::cerr) {
	return test_all<typename T::requirements>()(g, s);
      }
    };

    /// \brief Test axioms of a concept with local axioms
    template <typename T>
    struct test_all<T, true, true> {

      /// \brief Test axiom number "Index"
      template <typename Generator,
		typename... Axioms,
		typename Stream,
		typename Index = std::integral_constant<size_t, 0> >
      bool test(Generator& g, const std::tuple<Axioms...>& axioms,
		Stream& s, Index = Index()) {
	bool ret2 =
	  ::catsfoot::test(g, std::get<1>(std::get<Index::value>(axioms)),
			   type_to_string<T>() + "::" +
			   std::get<0>(std::get<Index::value>(axioms)), s);
	bool ret =
	  test(g, axioms, s, std::integral_constant<size_t, Index::value+1>());
	return ret && ret2;
      }

      /// \brief All axioms were tested
      template <typename Generator,
		typename... Axioms,
		typename Stream>
      bool test(Generator&, const std::tuple<Axioms...>&,
		Stream&, std::integral_constant<size_t, sizeof...(Axioms)>) {
	return true;
      }

      template <typename Generator,
		typename Stream = decltype(std::cerr)>
      bool operator()(Generator& g, Stream& s = std::cerr) {
	bool ret2 = test(g, T::get_axioms(), s);
	bool ret = test_all<typename T::requirements>()(g, s);
	return ret && ret2;
      }
    };

    /// \brief Test axioms of a list of requirements
    template <typename T, typename... U, bool B>
    struct test_all<concept_list<T, U...>, false, B> {
      template <typename Generator,
		typename Stream = decltype(std::cerr)>
      bool operator()(Generator& g, Stream& s = std::cerr) {
	bool ret = test_all<concept_list<U...> >()(g, s);
	bool ret2 = test_all<T>()(g, s);
	return ret && ret2;
      }
    };
  }

  /// \brief Tests all axioms of model Model and its sub-models.
  template <typename Model,
	    typename Generator,
	    typename Stream = decltype(std::cerr)>
  bool test_all(Generator& g, Stream& s = std::cerr) {
    return details::test_all<Model>()(g, s);
  }

}

#endif

