// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef __TYPE_MEMBER_HH
# define __TYPE_MEMBER_HH

namespace catsfoot {
  /// \brief Type used return by has_member_... in case member does not exist.
  struct undefined_member_type {};
}

/// \brief Generates has_member_##X to be a predicate and map for type member
///
/// has_member_##X<T>::value will be true if and only if T::X exists and is
/// a type.
///
/// has_member_##X<T>::member_type will be a typedef of T::X if it exists
/// and is a type. It will be catsfoot::undefined_type_member otherwise.
# define DEF_TYPE_MEMBER_PREDICATE(X...)				\
  namespace details {							\
    template <typename T>						\
    std::false_type has_member_##X##_helper				\
    (catsfoot::details::try_second, const T&) {				\
      static_assert(catsfoot::details::always_false<T>::value,		\
		    "Cannot use this function");			\
      return 0;								\
    }									\
    template <typename T,						\
	      typename = typename T::X>					\
    std::true_type has_member_##X##_helper				\
    (catsfoot::details::try_first, const T&) {				\
      static_assert(catsfoot::details::always_false<T>::value,		\
		    "Cannot use this function");			\
      return 0;								\
    }									\
    template <typename T>						\
    struct has_member_##X##_real_map {					\
      typedef typename T::X member_type;				\
    };									\
    template <typename>							\
    struct has_member_##X##_bad_map {					\
      typedef catsfoot::undefined_member_type member_type;		\
    };									\
  }									\
  template <typename T>							\
  struct has_member_##X:						\
    public catsfoot::identity						\
  <decltype(details::							\
	    has_member_##X##_helper					\
	    (catsfoot::details::try_first(), std::declval<T>()))>	\
  ::type {								\
    typedef typename catsfoot::identity					\
      <decltype(details::						\
		has_member_##X##_helper					\
		(catsfoot::details::try_first(), std::declval<T>()))>	\
      ::type								\
      super;								\
    typedef typename							\
      std::conditional<super::value,					\
		       details::has_member_##X##_real_map<T>,		\
		       details::has_member_##X##_bad_map<T> >		\
      ::type::member_type member_type;					\
  };									\
  template <>								\
  struct has_member_##X<void>:						\
    public std::false_type {						\
    typedef ::catsfoot::undefined_member_type member_type;		\
  };

#endif
