// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CALL_TUPLE_HH
# define CALL_TUPLE_HH

# include <tuple>
# include <type_traits>
# include "../type_traits/is_callable.hh"

namespace catsfoot {
  namespace details {
    template <typename Parameter, typename... Functions>
    struct call_tuple_helper {
      typedef
      std::tuple<typename std::decay
		 <typename is_callable
		  <const Functions(const Parameter&)>::result_type>
		 ::type...> return_type;

      static return_type
      call(const std::tuple<Functions...>&,
	   const Parameter&,
	   typename is_callable
	    <const Functions(const Parameter&)>::result_type... args) {
	return std::make_tuple(std::move(args)...);
      }

      template <typename... Args>
      static return_type
      call(const std::tuple<Functions...>& functions,
	   const Parameter& param,
	   Args... args) {
	return call
	  (functions, param,
	   std::move(args)...,
	   std::move(std::get<sizeof...(Args)>(functions)(param)));
      }
    };
  }

  /// \brief Calls a tuple of functions with an argument.
  ///
  /// \param functions The functions
  /// \param param The parameter for calling the function
  /// \returns the tuple of results
  template <typename Parameter, typename... Functions>
  std::tuple<typename std::decay
	     <typename is_callable
	      <const Functions(const Parameter&)>::result_type>
	     ::type...>
  call_tuple(const std::tuple<Functions...>& functions,
	     const Parameter& param) {
    return details::call_tuple_helper<Parameter, Functions...>
      ::call(functions, param);
  }

}

#endif
