// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef CALL_WITH_HH
# define CALL_WITH_HH

# include <tuple>
# include "../concept/concepts.hh"

namespace catsfoot {

  namespace details {
    template <typename T>
    struct add_const_ref {
      typedef const T& type;
    };

    template <typename T>
    struct add_ref {
      typedef T& type;
    };

    template <typename Op, typename Tuple>
    struct call_with_ret {
    };

    template <typename Op, typename... Args>
    struct call_with_ret<Op, std::tuple<Args...>&> {
      typedef
      decltype(std::declval<Op>()(std::declval<typename add_ref<Args>::type>()...))
      type;
      static const size_t size = sizeof...(Args);
    };

    template <typename Op, typename... Args>
    struct call_with_ret<Op, std::tuple<Args...> > {
      typedef
      decltype(std::declval<Op>()(std::declval<typename add_ref<Args>::type>()...))
      type;
      static const size_t size = sizeof...(Args);
    };

    template <typename Op, typename... Args>
    struct call_with_ret<Op, const std::tuple<Args...>&> {
      typedef
      decltype(std::declval<Op>()(std::declval<typename add_const_ref<Args>::type>()...))
      type;
      static const size_t size = sizeof...(Args);
    };

    template <typename Op, typename Tuple,
	      typename... Given,
	      typename = typename std::enable_if<(sizeof...(Given) == call_with_ret<Op, Tuple>::size)>::type>
    typename call_with_ret<Op, Tuple>::type
    call_with_it(Op&& op, Tuple&&, Given&&... args) {
      return std::forward<Op>(op)(std::forward<Given>(args)...);
    }

    template <typename Op, typename Tuple,
	      typename... OtherArgs,
	      typename = typename std::enable_if<(sizeof...(OtherArgs) != call_with_ret<Op, Tuple>::size)>::type,
	      typename = void>
    typename call_with_ret<Op, Tuple>::type
    call_with_it(Op&& op, Tuple&& args,
		 OtherArgs&&... otherargs) {
      return call_with_it(std::forward<Op>(op), std::forward<Tuple>(args),
			  std::forward<OtherArgs>(otherargs)...,
			  std::get<sizeof...(OtherArgs)>(args));
    }
  }

  /// \brief Calls a function with arguments from a \a std::tuple.
  ///
  /// \param op A function.
  /// \param args The tuple of arguments
  /// \returns the result of op(args...)
  template <typename Op, typename Tuple>
  typename details::call_with_ret<Op, Tuple>::type
  call_with(Op&& op, Tuple&& args) {
    return details::call_with_it(std::forward<Op>(op), std::forward<Tuple>(args));
  }

}

#endif
