// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <utils/black_box_equal.hh>
#include <iostream>
#include <dataset/dataset.hh>

struct foo {
private:
  int i;
  float j;
public:
  foo(int i, float j): i(i), j(j) {}
  foo(const foo&) = delete;

  int method_one() const {
    return i;
  }

  float method_two(const foo& other) const {
    return j + other.j;
  }

  int method_three(int max) const {
    return (i > max)?max:i;
  }
};

int main() {

 auto comparer = catsfoot::build_comparer<foo>()
   ( catsfoot::list_data_generator<int>
       ({0, 1, 2, -1, 41, 52}),
     std::function<int(const foo&)>([] (const foo& f) {
       return f.method_one();
       }),
     std::function<float(const foo&, const foo&)>
     ([] (const foo& f, const foo& other) {
	return f.method_two(other);
     }),
     std::function<int(const foo&, int)>
     ([] (const foo& f, int i) {
	return f.method_three(i);
     }));

  assert(!comparer(foo(0, 0.), foo(0, 3.)));
  assert(comparer(foo(0, 0.), foo(0, 0.)));
  assert(!comparer(foo(0, 0.), foo(1, 0.)));

  return 0;
}
