// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <concept/congruence.hh>
#include <dataset/dataset.hh>
#include <dataset/tuple_generator.hh>
#include <dataset/choose.hh>
#include <drivers/test_all_driver.hh>
#include <limits>

struct foo {
  bool operator()(int, int) const { return false; }
};

namespace catsfoot {
  template <>
  struct verified<congruence<foo, op_plus, int, int> >
    : public std::true_type{};
  template <>
  struct verified<equivalence<int, foo> >: public std::true_type {};
}

int main() {
  using namespace catsfoot;

  auto int_gen =
    list_data_generator<int>
    ({-1, 0, 1, 2, 3,
	std::numeric_limits<int>::min(),
	std::numeric_limits<int>::max()});

  auto generator =
    choose
    (int_gen,
     tuple_gen(int_gen),
     default_generator());

  assert_concept(congruence<foo, op_plus, int, int>());

  bool ret = !test_all<congruence<foo, op_plus, int, int> >(generator);

  ret = ret && test_all<congruence_eq<op_plus, int, int> >(generator);

  ret = ret && !check_unverified();

  return ret?0:1;
}
