// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#ifndef MONOID_HH
# define MONOID_HH

# include <concept/concepts.hh>
# include <axioms/axioms.hh>
# include <type_traits>

using namespace catsfoot;

template <typename T, typename Op, typename Id>
struct monoid: public concept {
  typedef concept_list<
    is_callable<Op(T, T)>,
    is_callable<Id>,
    std::is_convertible<typename is_callable<Op(T, T)>::result_type, T>,
    std::is_convertible<typename is_callable<Id(T)>::result_type, T>
    > requirements;

  static void associativity(const Op& op, const T& a,
		      const T& b, const T& c) {
    axiom_assert(op(a, op(b, c)) == op(op(a, b), c));
  }

  static void identity(const Op& op, const T& a, const Id& id) {
    axiom_assert((op(id(), a) == a) && (op(a, id()) == a));
  }

  AXIOMS(associativity, identity);
};

namespace catsfoot {
  template <>
  struct verified<monoid<int, op_plus, wrapped_constructor<int()> > >
    : public std::true_type
  {};
}

#endif
