// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <concept/product.hh>
#include <dataset/dataset.hh>
#include <dataset/tuple_generator.hh>
#include <dataset/choose.hh>
#include <drivers/test_all_driver.hh>
#include <dataset/random_term_generator.hh>

struct Foo {
  int a;
  int b;

  Foo(int a, int b): a(a), b(b) {}
  Foo(const Foo&) = default;

  bool operator==(const Foo& other) const {
    return (a == other.a) && (b == other.b);
  }
};

struct P1 {
  int operator()(const Foo& f) const {
    return f.a;
  }
};

struct P2 {
  int operator()(const Foo& f) const {
    return f.b;
  }
};

namespace catsfoot {
  template <>
  struct verified<simple_product<Foo, P1, P2> >:
    public std::true_type {};
}

int main() {
  std::mt19937 engine;
  auto term_generator =
    catsfoot::term_generator_builder<int, Foo>{}
  (engine,
   std::function<int()>([&engine] () {
       return std::uniform_int_distribution<int>()(engine); }),
   catsfoot::constructor<Foo(int, int)>());

  auto g = catsfoot::choose(term_generator,
			    catsfoot::default_generator{});

  auto g_and_tuples = catsfoot::choose(catsfoot::tuple_gen(g), g);

  catsfoot::test_all<catsfoot::simple_product<Foo, P1, P2> >(g_and_tuples);

  return 0;
}
