// This file is a part of Catsfoot.
// Copyright (C) 2011  Uni Research
//
// This program is free software: you can redistribute it and/or modify
// it under the terms of the GNU Lesser Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU Lesser Public License for more details.
//
// You should have received a copy of the GNU Lesser Public License
// along with this program.  If not, see <http://www.gnu.org/licenses/>.

#include <type_traits/is_callable.hh>
#include <type_traits/is_constructible.hh>
#include <type_traits/type_member.hh>

DEF_TYPE_MEMBER_PREDICATE(memb);

struct a {
  void operator()(int) {}
};

struct b {
  void operator()(int) const {}
};

struct c {
  typedef int memb;
  c(const c&) = delete;
};

struct d: public c {
};

struct e {
  typedef void memb;
  e(const d&) {}
};

struct f {
  void operator()(int, int) const {
  }

  int operator()(int) const {
    return 0;
  }
};



int main() {
  static_assert(catsfoot::is_callable<f(int, int)>::value,
		"Test failed");

  static_assert(catsfoot::is_constructible<int()>::value,
		"Test failed");

  static_assert(catsfoot::is_callable<a(int)>::value,
		"Test failed");
  static_assert(!catsfoot::is_callable<const a(int)>::value,
		"Test failed");
  static_assert(catsfoot::is_callable<b(int)>::value,
		"Test failed");
  static_assert(catsfoot::is_callable<const b(int)>::value,
		"Test failed");

  static_assert(catsfoot::is_constructible<e(const d&)>::value,
		"Test failed");

  static_assert(!catsfoot::is_constructible<c(int)>::value,
		"Test failed");

  static_assert(!catsfoot::is_constructible<c(d)>::value,
		"Test failed");

  static_assert(has_member_memb<c>::value,
		"Test failed");

  static_assert(!has_member_memb<f>::value,
		"Test failed");

  static_assert(std::is_same<typename has_member_memb<f>::member_type,
		catsfoot::undefined_member_type>::value,
		"Test failed");

  static_assert(std::is_same<typename has_member_memb<c>::member_type,
		int>::value,
		"Test failed");

  static_assert(std::is_same<typename has_member_memb<e>::member_type,
		void>::value,
		"Test failed");

  return 0;
}
